#include "StdMpi.h"

namespace AMDiS {

  int intSizeOf(int &data)
  {
    return 1;
  }

  int intSizeOf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data)
  {
    return data.size() * 2;
  }

  int intSizeOf(std::vector<MeshStructure> &data)
  {
    int s = 0;
    for (unsigned int i = 0; i < data.size(); i++)
      s += data[i].getCode().size() + 2;

    return s;
  }

  int intSizeOf(std::vector<WorldVector<double> > &data)
  {
    return data.size() * Global::getGeo(WORLD);
  }

  int intSizeOf(std::vector<int> &data)
  {
    return data.size();
  }

  int intSizeOf(std::vector<double> &data)
  {
    return data.size();
  }
  
  int intSizeOf(std::vector<std::pair<int, int> > &data)
  {
    return data.size() * 2;
  }

  int intSizeOf(std::vector<AtomicBoundary> &data)
  {
    return data.size() * 3;
  }

  int intSizeOf(std::vector<BoundaryObject> &data)
  {
    return data.size() * 5;
  }

  int intSizeOf(std::vector<const DegreeOfFreedom*> &data)
  {
    return data.size();
  }

  void makeBuf(int &data, int *buf)
  {
    buf[0] = data;
  }

  void makeFromBuf(int &data, int *buf, int bufSize)
  {
    data = buf[0];
  }

  void makeBuf(std::map<const DegreeOfFreedom*, DegreeOfFreedom> &data, int* buf)
  {
    int i = 0;
    for (std::map<const DegreeOfFreedom*, DegreeOfFreedom>::iterator it = data.begin();
	 it != data.end(); ++it) {
      buf[i++] = *(it->first);
      buf[i++] = it->second;
    }
  }
  
  void makeFromBuf(std::vector<std::pair<DegreeOfFreedom, DegreeOfFreedom> > &data, 
		   int *buf, int bufSize)
  {    
    if (bufSize == 0)
      return;

    TEST_EXIT(bufSize % 2 == 0)("This should not happen!\n");

    data.clear();
    data.reserve(bufSize / 2);

    for (int i = 0; i < (bufSize / 2); i++)
      data.push_back(std::make_pair(buf[i * 2], buf[i * 2 + 1]));
  }

  void makeBuf(std::vector<MeshStructure> &data, unsigned long int *buf)
  {    
    int pos = 0;
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[pos++] = data[i].getCode().size();
      buf[pos++] = data[i].getNumElements();

      for (unsigned int j = 0; j < data[i].getCode().size(); j++)
	buf[pos++] = data[i].getCode()[j];
    }
  }

  void makeFromBuf(std::vector<MeshStructure> &data, 
		   unsigned long int *buf, int bufSize)
  {
    int pos = 0;

    while (pos < bufSize) {       
      int codeSize = buf[pos++];
      int nElements = buf[pos++];
      std::vector<unsigned long int> code;
      code.resize(codeSize);
      for (int i = 0; i < codeSize; i++)
	code[i] = buf[pos++];

      MeshStructure meshCode;
      meshCode.init(code, nElements);

      data.push_back(meshCode);
    }	
  }

  void makeBuf(std::vector<WorldVector<double> > &data, double *buf)
  {
    int dimOfWorld = Global::getGeo(WORLD);
    int pos = 0;
    for (unsigned int i = 0; i < data.size(); i++)
      for (int j = 0; j < dimOfWorld; j++)
	buf[pos++] = data[i][j];
  }

  void makeFromBuf(std::vector<WorldVector<double> > &data, double *buf, int bufSize)
  {
    int dimOfWorld = Global::getGeo(WORLD);
    TEST_EXIT(bufSize % Global::getGeo(WORLD) == 0)("This should not happen!\n");

    int pos = 0;
    data.resize(bufSize / Global::getGeo(WORLD));
    for (unsigned int i = 0; i < data.size(); i++)
      for (int j = 0; j < dimOfWorld; j++)
	data[i][j] = buf[pos++];
  }

  void makeBuf(std::vector<int> &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++)
      buf[i] = data[i];
  }

  void makeFromBuf(std::vector<int> &data, int *buf, int bufSize)
  {
    data.resize(bufSize);

    for (int i = 0; i < bufSize; i++)
      data[i] = buf[i];
  }

  void makeBuf(std::vector<double> &data, double *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++)
      buf[i] = data[i];
  }

  void makeFromBuf(std::vector<double> &data, double *buf, int bufSize)
  {
    data.resize(bufSize);

    for (int i = 0; i < bufSize; i++)
      data[i] = buf[i];
  }

  void makeBuf(std::vector<std::pair<int, int> > &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[i * 2] = data[i].first;
      buf[i * 2 + 1] = data[i].second;
    }
  }

  void makeBuf(std::vector<AtomicBoundary> &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[i * 3] = data[i].rankObj.elIndex;
      buf[i * 3 + 1] = data[i].rankObj.subObj;
      buf[i * 3 + 2] = data[i].rankObj.ithObj;
    }
  }

  void makeFromBuf(std::vector<AtomicBoundary> &data, int *buf, int bufSize)
  {
    if (bufSize == 0)
      return;

    TEST_EXIT(bufSize % 3 == 0)("This should not happen!\n");    

    data.resize(bufSize / 3);
    for (int i = 0; i < bufSize / 3; i++) {
      data[i].rankObj.elIndex = buf[i * 3];
      data[i].rankObj.subObj = static_cast<GeoIndex>(buf[i * 3 + 1]);
      data[i].rankObj.ithObj = buf[i * 3 + 2];
    }
  }

  void makeBuf(std::vector<BoundaryObject> &data, int *buf)
  {
    for (unsigned int i = 0; i < data.size(); i++) {
      buf[i * 5] = data[i].elIndex;
      buf[i * 5 + 1] = data[i].elType;
      buf[i * 5 + 2] = data[i].subObj;
      buf[i * 5 + 3] = data[i].ithObj;
      buf[i * 5 + 4] = data[i].reverseMode;
    }
  }

  void makeFromBuf(std::vector<BoundaryObject> &data, int *buf, int bufSize)
  {
    if (bufSize == 0)
      return;

    TEST_EXIT(bufSize % 5 == 0)("This should not happen!\n");    

    data.resize(bufSize / 5);
    for (int i = 0; i < bufSize / 5; i++) {
      data[i].elIndex = buf[i * 5];
      data[i].elType = buf[i * 5 + 1];
      data[i].subObj = static_cast<GeoIndex>(buf[i * 5 + 2]);
      data[i].ithObj = buf[i * 5 + 3];
      data[i].reverseMode = static_cast<bool>(buf[i * 5 + 4]);
    }
  }

}
